import numpy as np
from sklearn.metrics import (
    f1_score,
    precision_score,
    recall_score,
    accuracy_score,
    matthews_corrcoef,
    auc,
    roc_curve,
    precision_recall_curve,
)


class Result:
    def __init__(self) -> None:
        self.auc = 0
        self.acc = 0
        self.f1 = 0
        self.aupr = 0
        self.loss = 0

    def get_result(self):
        return {
            "AUC": self.auc,
            "ACC": self.acc,
            "F1 Score": self.f1,
            "AUPR": self.aupr,
            "Loss": self.loss,
        }

    def add(self, result):
        self.auc = self.auc + result.auc
        self.acc = self.acc + result.acc
        self.f1 = self.f1 + result.f1
        self.aupr = self.aupr + result.aupr
        self.loss = self.loss + result.loss

    def divide(self, k):
        self.auc = self.auc / k
        self.acc = self.acc / k
        self.f1 = self.f1 / k
        self.aupr = self.aupr / k
        self.loss = self.loss / k


def get_prediction_results(y_test, y_predict, threshold):
    result = Result()

    binary_predict = np.where(np.array(y_predict) >= threshold, 1, 0).tolist()

    result.acc = accuracy_score(y_test, binary_predict)
    result.f1 = f1_score(y_test, binary_predict, average="macro", zero_division=1)
    result.recall = recall_score(
        y_test, binary_predict, average="macro", zero_division=1
    )
    result.precision = precision_score(
        y_test, binary_predict, average="macro", zero_division=1
    )
    result.mcc = matthews_corrcoef(y_test, binary_predict)

    precision, recall, thresholds = precision_recall_curve(y_test, y_predict)
    numerator = 2 * recall * precision
    denominator = recall + precision
    f1_scores = np.divide(
        numerator, denominator, out=np.zeros_like(denominator), where=(denominator != 0)
    )
    result.max_f1 = np.max(f1_scores)

    fpr, tpr, thresholds = roc_curve(y_test, y_predict, pos_label=1)
    result.auc = auc(fpr, tpr)
    result.tpr = tpr
    result.fpr = fpr

    return result



from .basemodel import BaseModel, ModelFactory
from .config import ModelConfig, OptimizerConfig, MethodConfig
from .data import Data, TrainTestSpliter, SimplePytorchData, SimplePytorchDataTrainTestSpliter
from .evaluation import Result, get_prediction_results
from .methods import FeatureExtractor
from .optimization import SimpleTester, SimpleTrainer
from .optimization import Tester, Trainer, cross_validation
from .optimization import backpropagation, predict_error



import abc


class Config(abc.ABC):
    @abc.abstractmethod
    def get_configuration(self):
        pass

    @abc.abstractmethod
    def get_summary(self):
        pass


class ModelConfig(Config):
    def __init__(self):
        super().__init__()
        self.model_name = None
        self.embedding_dim = None  # dimension of embedding
        self.device = None

    def get_configuration(self):
        return {
            "model_name": self.model_name,
            "embed_dim": self.embedding_dim,
        }

    def get_summary(self):
        return {
            "model_name": self.model_name,
            "embed_dim": self.embedding_dim,
        }


class MethodConfig(Config):
    def __init__(self):
        super().__init__()
        self.method_name = None

    def get_configuration(self):
        return {
            "method_name": self.method_name,
        }

    def get_summary(self):
        return {
            "method_name": self.method_name,
        }


class OptimizerConfig(Config):
    def __init__(self) -> None:
        super().__init__()
        self.optimizer = None
        self.criterion = None
        self.lr = None  # learning rate
        self.batch_size = None
        self.n_epoch = None
        self.exp_name = None
        self.save = False
        self.save_path = None
        self.device = 'cpu'
        self.report_size = 100  # batch to report ratio
        self.threshold = 0.5

    def get_configuration(self):
        return {
            "optimizer": self.optimizer,
            "lr": self.lr,
            "batch_size": self.batch_size,
            "n_epoch": self.n_epoch,
        }

    def get_summary(self):
        return {
            "optimizer": self.optimizer,
            "lr": self.lr,
            "batch_size": self.batch_size,
            "n_epoch": self.n_epoch,
        }



import abc
import torch

from .utils import logging as base_logger

logger = base_logger.getLogger(__name__)


class Data(abc.ABC):
    def __init__(self, **kwargs) -> None:
        super().__init__()

    @abc.abstractmethod
    def extend(self, **kwargs):
        raise NotImplementedError

    @abc.abstractmethod
    def subset(self, indices) -> tuple:
        raise NotImplementedError


class TrainTestSpliter(abc.ABC):

    def __init__(self, k):
        self.k = k

    @abc.abstractmethod
    def split(self, i):  # Return Train and Test Data
        logger.info(f'splitting {i}th fold')
        raise NotImplementedError


class SimplePytorchData(Data):
    def __init__(self, X: torch.Tensor, y: torch.Tensor, **kwargs) -> None:
        super().__init__(**kwargs)
        logger.info(f'Initializing SimplePytorchData with X shape : {X.shape} and y shape : {y.shape}')
        self.X = X
        self.y = y

    def extend(self, X: torch.Tensor, y: torch.Tensor):
        if self.X is None or self.y is None:
            self.X = X
            self.y = y
        else:
            self.X = torch.cat((self.X, X), 0)
            self.y = torch.cat((self.y, y), 0)

    def subset(self, indices) -> tuple:
        return self.y[indices], self.y[indices]


class SimplePytorchDataTrainTestSpliter(TrainTestSpliter):

    def __init__(self, k, simple_data):
        super().__init__(k)
        logger.info(f'Initializing SimplePytorchDataTrainTestSpliter')
        self.X = simple_data.X
        self.y = simple_data.y

        self.data_size = simple_data.X.shape[0]

        subsets = dict()
        subset_size = int(self.data_size / self.k)
        remain = set(range(0, self.data_size))
        for i in range(self.k - 1):
            subsets[i] = random.sample(remain, subset_size)
            remain = remain.difference(subsets[i])
        subsets[k - 1] = remain

        self.subsets = subsets

    def split(self, i):
        indices = set(range(0, self.data_size))
        test_indices = list(self.subsets[i])
        train_indices = list(indices.difference(self.subsets[i]))

        train_data = SimplePytorchData(self.X[train_indices], self.y[train_indices])
        test_data = SimplePytorchData(self.X[test_indices], self.y[test_indices])
        return train_data, test_data



import abc
import os.path
import random
import torch
from torch.utils.data import TensorDataset, DataLoader

from .basemodel import BaseModel, ModelFactory
from .config import OptimizerConfig
from .data import Data, TrainTestSpliter, SimplePytorchData
from .evaluation import Result, get_prediction_results
from .utils import logging as base_logger

logger = base_logger.getLogger(__name__)


class Trainer(abc.ABC):
    @abc.abstractmethod
    def train(self, model: BaseModel, data: Data, config: OptimizerConfig) -> Result:
        raise NotImplementedError


class Tester(abc.ABC):
    @abc.abstractmethod
    def test(self, model: BaseModel, data: Data, config: OptimizerConfig) -> Result:
        raise NotImplementedError


def cross_validation(train_test_spliter: TrainTestSpliter, model_factory: ModelFactory, trainer: Trainer,
                     tester: Tester, config: OptimizerConfig) -> Result:
    k = train_test_spliter.k
    logger.info(f'Start {k}-fold Cross Validation with config : {config.exp_name}')

    result = Result()
    for i in range(k):
        logger.info(f"---- Fold {i + 1} ----")

        train_data, test_data = train_test_spliter.split(i)

        model = model_factory.make_model()
        trainer.train(model=model, data=train_data, config=config)
        test_result = tester.test(model=model, data=test_data, config=config)
        logger.info(f'Result of fold {i + 1} : {test_result.get_result()}')
        model.destroy()

        result.add(test_result)

    result.divide(k=k)

    logger.info(
        f'{k}-fold result: avg_auc: {result.auc}, avg_acc: {result.acc}, avg_f1: {result.f1}, avg_aupr: {result.aupr}')

    return result


def _batch_optimize(loader, model, config: OptimizerConfig, ):
    model.classifier.train()
    optimizer = config.optimizer(model.classifier.parameters(), lr=config.lr)

    for epoch in range(config.n_epoch):
        running_loss = 0.0
        for j, data in enumerate(loader, 0):
            X, y = data

            loss, running_loss = predict_error(X, config.criterion, model, running_loss, y)
            backpropagation(loss, optimizer)

            if j % config.report_size == config.report_size - 1:
                loss = running_loss / (config.report_size * config.batch_size)
                logger.info(f'loss: {loss:.4f}    [{epoch + 1}, {j + 1:5d}]')
                running_loss = 0

        if config.save:
            if epoch % 5 == 0 or epoch == config.n_epoch - 1:
                m = os.path.join(config.save_path, model.model_config.model_name + ".pth")
                torch.save(model.classifier.state_dict(), m)


def predict_error(X, loss_function, model, running_loss, y):
    pred = model.classifier(X)
    loss = loss_function(pred, y)
    running_loss += loss.item()
    return loss, running_loss


def backpropagation(loss, optimizer):
    optimizer.zero_grad()
    loss.backward()
    optimizer.step()


def _evaluate(model: BaseModel, loader, config: OptimizerConfig):
    model.classifier.eval()
    total_labels = []
    total_predictions = []
    total_loss = 0.0
    for data in loader:
        inputs, labels = data
        outputs = model.predict(inputs)
        total_loss += config.criterion(outputs, labels).item()
        total_labels.extend(labels.cpu().numpy())
        total_predictions.extend(torch.sigmoid(outputs).cpu().numpy())
    result = get_prediction_results(total_labels, total_predictions, config.threshold)
    result.loss = total_loss / len(loader)
    return result


class SimpleTrainer(Trainer):
    def train(self, model: BaseModel, data: SimplePytorchData, config: OptimizerConfig) -> Result:
        logger.info(f'Running Simple Trainer with config : {config.exp_name}')

        logger.info(f'moving data and model to {config.device}')
        model.classifier = model.classifier.to(config.device)
        dataset = TensorDataset(data.X.to(config.device), data.y.to(config.device))
        loader = DataLoader(dataset, batch_size=config.batch_size, shuffle=True)

        _batch_optimize(loader, model, config)
        result = _evaluate(model, loader, config)

        logger.info(f'Result on Train Data : {result.get_result()}')
        return result


class SimpleTester(Tester):
    def test(self, model: BaseModel, data: SimplePytorchData, config: OptimizerConfig = None) -> Result:
        logger.info(f'Running Simple Tester with config : {config.exp_name}')

        logger.info(f'moving data and model to {config.device}')
        model.classifier = model.classifier.to(config.device)
        dataset = TensorDataset(data.X.to(config.device), data.y.to(config.device))
        loader = DataLoader(dataset, batch_size=config.batch_size, shuffle=True)
        result = _evaluate(model, loader, config)

        logger.info(f'Result on Test Data : {result.get_result()}')
        return result



import abc

from .utils import logging as base_logger

logger = base_logger.getLogger(__name__)


class BaseModel(abc.ABC):
    def __init__(self, model_config) -> None:
        self.model_config = model_config
        self.model = None

    @abc.abstractmethod
    def destroy(self):
        logger.info(f'Model {self.model} has been deleted')
        raise NotImplementedError

    @abc.abstractmethod
    def predict(self, **kwargs):
        raise NotImplementedError

    @abc.abstractmethod
    def summary(self):
        raise NotImplementedError


class ModelFactory(abc.ABC):
    @abc.abstractmethod
    def make_model(self) -> BaseModel:
        raise NotImplementedError



import abc

from .utils import logging as base_logger

logger = base_logger.getLogger(__name__)


class FeatureExtractor(abc.ABC):

    @abc.abstractmethod
    def build(self, **kwargs):
        logger.info(f'Building Feature Extractor')
        raise NotImplementedError
